#!/usr/bin/perl

# Input:
#  - location of actual mpirun command
#  - command line arguments as meant for mpirun
#
# Result: exec mpirun with modified command line

use File::Temp qw/ tempfile /;

main();

sub main {
    $mpirun = shift @ARGV;
    identify_mpirun_native_options();

    @record_env_list = ();
    @record_env_propagate = ();
    %extra_env_from_cmdline = ();

    if ($ENV{'MPI_ROOT'} ne '') { record_env('MPI_ROOT'); }
    if ($ENV{'OPAL_PREFIX'} ne '') { record_env('OPAL_PREFIX'); }
    if ($ENV{'OPAL_LIBDIR'} ne '') { record_env('OPAL_LIBDIR'); }
    if ($ENV{'PMIX_INSTALL_PREFIX'} ne '') {
        record_env('PMIX_INSTALL_PREFIX');
    }
    if ($ENV{'HWLOC_PLUGINS_PATH'} ne '') {
        record_env('HWLOC_PLUGINS_PATH');
    }

    # handle -help and --help specially:
    # run the base mpirun -help, then list our Platform options.
    # stock/mpirun now has optional categories of help so pass that on too
    $help = 0;
    if ($ARGV[0] eq '-help' || $ARGV[0] eq '--help') {
        $help = 1;
        if (defined($ARGV[1])) {
            system("$mpirun -help $ARGV[1]");
        } else {
            system("$mpirun -help full");
        }
        print "\nExtra options from Spectrum-MPI " .
            "(that translate to similar Open MPI options):\n";
    }

    initialize_state_vars();
    if ($help) { exit(0); }

    read_smpi_confs();
    process_argv();
    # Now $force_pml is set, $prefer_pml is no longer really used.
    produce_new_arguments();
    @cmd = ($mpirun, @argv_additions, @argv_passthrough);

    if ($use_jsrun) {
        # use jsrun from $MPI_ROOT/jsm_pmix/bin/jsrun of MPI_JSRUN
        $jsrun = $ENV{'MPI_ROOT'}."/jsm_pmix/bin/jsrun";
        if (defined $ENV{'JSM_ROOT'}) {
            $jsrun = $ENV{'JSM_ROOT'}."/bin/jsrun";
        }
        if( ! -e $jsrun && $run) {
            print "Error: Cannot find the jsrun binary.\nPath: $jsrun\n";
            exit(1);
        }

        set_extra_env_for_jsrun();
        check_for_unrecognized_mpirun_args(\@argv_passthrough, 0,
            "cmdline arg");
        check_for_unrecognized_mpirun_args(\@argv_additions, 0,
            "generated arg");

        @cmd = ($jsrun, @jsrun_args);
    }

    if ($show || $env_output_file ne '') {
        for $var (@record_env_list) {
            display_env_setting($var);
        }
    }
    if (scalar(@record_env_propagate)) {
        $val = join(",", @record_env_propagate);
        $ENV{'OMPI_MCA_mca_base_env_list_distro'} = $val;
        if ($show || $env_output_file ne '') {
            display_env_setting("OMPI_MCA_mca_base_env_list_distro");
        }
    }
    if ($env_output_file_fh_isopen) {
        close($env_output_file_fh);
        if (! $use_jsrun) {
            print "Environment written to $env_output_file\n";
        }
    }
    if ($show) {
        print "\n";
    }

    if ($show) {
        @esc_cmd = ();
        for $arg (@cmd) {
            push @esc_cmd, myescape($arg);
        }
        print join(" ", @esc_cmd), "\n";
    }
    if ($run) {
        exec(@cmd);
    }
}


# Use "mpirun -help" to look for <arg0> and <arg1> to identify
# options like
#     -np <number>
#     --mca <var> <value>
# that show argv options that need to be skipped over.
# Result goes into globals @skip1 and @skip2

sub identify_mpirun_native_options {
    my $line;
    my $number_of_options;
    my @list;
    my %tmp;

    @skip1 = ();
    @skip2 = ();
    open(FP, "$mpirun -help parsable 2>/dev/null |");
    while ($line = <FP>) {
        chomp $line;

        $line =~ /^(.):([^:]*):([^:]*):([0-9]*):.*/;
        $number_of_options = $4;
        @list = ();  # arguments represented by parts $1 $2 $3
        if ($1 ne '0') { push(@list, "-$1", "--$1"); }
        if ($2 ne '(null)') { push(@list, "-$2", "--$2"); }
        if ($3 ne '(null)') { push(@list, "-$3", "--$3"); }
# The -help output isn't accurate with regard to whether an option allows
# one or two dashes, so the above represents reality more than if we
# interpreted the -help output strictly.

        if ($number_of_options == 2) {
            push(@skip2, @list);
        }
        elsif ($number_of_options == 1) {
            push(@skip1, @list);
        }
    }

# They don't really need to be unique, but might as well make it so:
    %tmp = (); $tmp{$_} = 1 for @skip1; @skip1 = sort keys(%tmp);
    %tmp = (); $tmp{$_} = 1 for @skip2; @skip2 = sort keys(%tmp);

# defaults that should be good enough to get by if something goes wrong
# in the above:
    if (!(scalar @skip1) && !(scalar @skip2)) {
        @skip1 = ('--H', '--N', '--am', '--app', '--bind-to', '--c', '--cartofile', '--cf', '--cpu-list', '--cpu-set', '--cpus-per-proc', '--cpus-per-rank', '--debugger', '--default-hostfile', '--h', '--help', '--hnp', '--host', '--hostfile', '--launch-agent', '--machinefile', '--map-by', '--max-restarts', '--max-vm-size', '--n', '--np', '--npernode', '--npersocket', '--ompi-server', '--output-filename', '--path', '--personality', '--ppr', '--prefix', '--preload-files', '--rank-by', '--rankfile', '--report-events', '--report-pid', '--report-uri', '--rf', '--stdin', '--timeout', '--tune', '--wd', '--wdir', '--x', '--xml-file', '--xterm', '-H', '-N', '-am', '-app', '-bind-to', '-c', '-cartofile', '-cf', '-cpu-list', '-cpu-set', '-cpus-per-proc', '-cpus-per-rank', '-debugger', '-default-hostfile', '-h', '-help', '-hnp', '-host', '-hostfile', '-launch-agent', '-machinefile', '-map-by', '-max-restarts', '-max-vm-size', '-n', '-np', '-npernode', '-npersocket', '-ompi-server', '-output-filename', '-path', '-personality', '-ppr', '-prefix', '-preload-files', '-rank-by', '-rankfile', '-report-events', '-report-pid', '-report-uri', '-rf', '-stdin', '-timeout', '-tune', '-wd', '-wdir', '-x', '-xml-file', '-xterm');
        @skip2 = ('-mca', '-gmca', '--mca', '--gmca');
    }

    close(FP);
}

# Initialize state vars for Platform options
# Also mix the state vars with documentation for use with -help / --help
sub initialize_state_vars {

# ------------------------------------------------
# -------- -hostlist
    @hostlist = (); # will be stored as host,host,host,..

    if (0 && $help) { print << 'EOF'

  [Alternate launching method]
    -hostlist <list>  : takes a comma separated list of hosts, each with
                        an optional colon and a number, for example
                            -hostlist hostA:8,hostB:8,hostC
                        and lauches ranks in the specified layout.
                        Note: this feature usurps the mapping stage of affinity
                        so most affinity options are not available with
                        -hostlist. Options that are compatible with -hostlist
                        include
                            --bind-to core
                            --map-by slot --bind-to socket
                            --map-by slot:pe=n
                        If these basic affinity options aren't adequate, a
                        different launching method should be used.
EOF
    }


# ------------------------------------------------
# -------- interconnect selection
    $prefer_pml = '';
    $force_pml = '';
    $enable_hcoll = 0;
    $hcoll_priority = -1;
    $allow_cm_mtl = '';
    $prefer_cm_mtl = '';
    $use_pami_verbsbypass = -1;
    $use_pami_noib = 0;
    $use_pami_ft = 0; #modifies other pami's specified on command line.
    $use_pami = 0; # some pami flag set on cmdline (overrides conf file)
    %allow_ob1_btl = ();  # Hash, keys represent list of btls to be active.
    %allow_ob1_btl_forshmem = ();
                          # Empty lists here mean allow all btls.
                          # If 'nic' is specified, vader/shm are removed.
    @all_btls_offhost = ('tcp', 'usnic', 'openib');
    @all_btls_onhost = ('vader');
    $uname_m = `uname -m`; chomp $uname_m;
#   -OFI / -ofi    : use OpenFabrics OFI via the PML cm and the MTL ofi
    if ($help) { print << 'EOF'

  [Interconnect selection]
    -PAMI / -pami  : use IBM PAMI via the PML pami
    -MXM / -mxm    : use Mellanox MXM via the PML yalla (the preferred method)
    -TCP / -tcp    : use TCP/IP via the PML ob1 and the BTL tcp
    -IBV / -ibv    : use OpenFabrics infiniband via the
                     PML ob1 and the BTL openib
                       aliases: -ib / -openib
EOF
# We've removed (D136551): -MXMC / -mxmc  : use Mellanox MXM via the PML cm and the MTL mxm
    }
    if (($help) && ($uname_m eq 'x86_64')) { print << 'EOF'
    -PSM / -psm    : use QLogic/Intel PSM via the PML cm and the MTL psm
    -PSM2 / -psm2  : use Intel Omni-Path HFI via the PML cm and the MTL psm2
    -USNIC / -usnic : use usNIC via PML ob1 and BTL usnic
EOF
    }
    if ($help) { print << 'EOF'

    In all of the above the capital option equates to forcing the specified
    PML / MTL / BTL, and the lower case option only equates to specifying a
    higher priority for the selected interconnect.
EOF
    }

# ------------------------------------------------
# -------- PAMI additional options
    if ($help) { print << 'EOF'

  [Additional PAMI options]
    -verbsbypass <ver> : use PAMI's verbs bypass.  <ver> reflects
                         Mellanox OFED version installed:
                            (ver= 4.1, auto, and off)
                         auto find out the installed compatible MOFED version on the mpirun node
                         (auto assumes complete cluster installed with same MOFED level)
    -pami_noib         : use PAMI on a single node with no Infiniband card. [ppc64 only]
EOF
    }

    if ($help) { print << 'EOF'

  [On-host communication method]
    -intra nic     : use the off-host BTL for on-host traffic as well
    -intra vader   : use BTL=vader (shared memory) for on-host traffic
                     (only applies if the PML is already ob1)
    -intra shm     : equivalent to -intra=vader
EOF
    }

# ------------------------------------------------
# -------- -prot -protlazy (-prot for init, -protlazy for finalize)
    $use_prot = 0;  # 1 = regular prot in init, 2 = lazy prot in finalize

    if ($help) { print << 'EOF'

  [Display interconnect]
    -prot          : display a table of what interconnect type each host uses
                     (first rank on each host connects to all peer hosts to
                     establish connections that might otherwise be on-demand)
    -protlazy      : less aggressive version of -prot that runs at finalize
                     and without establishing connections, so many peers
                     might be unconnected.
EOF
    }

# ------------------------------------------------
# -------- -jsrun
    $use_jsrun = 0;
    @jsrun_args = ();
    @jsrun_ld_library_path_additions = ();

    if (0 && $help) { print << 'EOF'

  [jsrun]
    -jsrun <args>  : launch with the jsrun command. Everything beyond the
                     -jsrun argument is passed directly to jsrun without
                     any parsing by mpirun. By default the jsrun executable
                     used is $MPI_ROOT/jsm_pmix/bin/jsrun. The environment
                     variable JSM_ROOT can be used to specify an alternate
                     path to the JSM installation directory. In which case
                     the jsrun executable used is $JSM_ROOT/bin/jsrun.
EOF
    }

#
# Disabling HCOLL for IBM Spectrum MPI v10.1, and hope to re-enable for
# a future release.
# 
# ------------------------------------------------
# -------- Collective Offloading
#    if ($help) { print << 'EOF'
#
#  [Collective Offloading]
#    -HCOLL / -FCA  : use Mellanox HCOLL collective component (aka FCA 3.x)
#    -hcoll / -fca  : use Mellanox HCOLL collective component only if coll_ibm
#                     chooses not to satisfy the collective call.
#EOF
#    }


# ------------------------------------------------
# -------- -stdio=
    $stdio_tag_output = 0;
    $stdio_timestamp_output = 0;
    $stdin_goes_to = ''; # rank / all / none
    $output_filename = '';

    if ($help) { print << 'EOF'

  [Stdio options]
    -stdio p                       : prefix each rank's output with [job,rank]
    -stdio t                       : add timestamp to output
    -stdio i[+|all|-|none|<rank>]  : send stdin to all ranks (+), no ranks (-)
                                     or a single specific rank
    -stdio file:prefix             : send output to files named <prefix>.<rank>
    -stdio <option>,<option>,..    : comma separated list of the above options
EOF
    }


# ------------------------------------------------
# -------- -netaddr mpirun:ip/n,rank:ip/n
$netaddr = '';
$netaddr_mpirun = '';
$netaddr_rank = '';

    if ($help) { print << 'EOF'

  [IP network selection]
    -netaddr <spec>,<spec>,..         :  specify what network(s) to use for
                                         IP traffic. This option applies
                                         to both control messages, and the
                                         regular MPI rank traffic
    -netaddr <type>:<spec>,<spec>,..  :  individually specify the networks
                                         for different types of traffic
    <type> can be any of
        rank       :  specify network for regular MPI rank-to-rank traffic
        control    :  specify network for control messages, eg launching
        mpirun     :  synonym for "control"
    <spec> can be either
        interface name  :  eg eth0 or ib0 etc
        CIDR notation   :  eg 10.10.1.0/24
EOF
    }


# ------------------------------------------------
# -------- -restrict_libs=
    $restrict_libs = "";

    if ($help) { print << 'EOF'

  [libnl / libnl3 collision avoidance]
    -restrict_libs nl / libnl /         : only load libraries compatible with
                   ^nl3 / ^libnl3         libnl (skip libnl3)
    -restrict_libs nl3 / libnl3 /       : only load libraries compatible with
                   ^nl / ^libnl           libnl3 (skip libnl)
    -restrict_libs consistent           : reject inconsistency vs current state
    -restrict_libs none                 : no restrictions
    -restrict_libs default              : "consistent"
    -restrict_libs v / vv / vvv         : print a message when rejecting an MCA
    -restrict_libs <option>,<option>,.. : comma separated list of the above
                                          options
    The levels of verboseness are
      v : print a message when rejecting a library due to a libnl conflict
      vv : for each library that uses libnl/nl3 print what it was detected as
      vvv : for every library print what it was detected as
EOF
    }


# ------------------------------------------------
# -------- -aff

# Most of the advanced Open MPI affinity-related options are dangerous
# in the sense that they affect rank-to-host mapping which can be
# counterintuitive. So here we only use very simple options.

# Platform options to support:
#   v / vv
#   cycle:[ hwthread / core / socket (default) / numa / board ]
#   bandwidth (also "default"/"auto[matic]"/"on")
#   latency
#   width:[ hwthread / core / socket (default) / numa / board ]

$use_aff = 0;
$aff_v = 0;
$bind_to = 'core'; # default if aff is turned on and a width isn't specified
$map_by = 'socket';
$rank_by = 'core';

# Translations:
#
# bandwidth:
#   map by socket, bind to width, rank by width
# cycle:unit (like bandwidth but the leave ordering interleaved)
#   map by unit, bind to width, rank by unit
# latency:
#   map by width, bind to width, rank by width

    if ($help) { print << 'EOF'

  [Affinity options]
    -aff on                 : turn on affinity with default option (bandwidth)
    -aff off                : turn off affinity (unbind)
    -aff v / -aff vv        : verbose
    -aff bandwidth          : interleave sockets, use natural hardware order
    -aff latency            : pack ranks across the natural hardware order
    -aff cycle:<unit>       : interleave binding over the specified element
    -aff width:<unit>       : bind each rank to an element of this size
         <unit> can be hwthread, core, socket, numa, or board.

    -aff default               : same as "bandwidth" above
    -aff auto[matic]           : same as "bandwidth" above
    -aff none                  : same as "off" above
    -aff <option>,<option>,..  : comma separated list of the above options
EOF
    }

# ------------------------------------------------
# -------- -gpu
    $use_gpu = 0;
    $disable_gpu_hooks = 0; # Want hooks on Always! Independant of use_gpu.
    if ($help) { print << 'EOF'

  [GPU support]
    -gpu           : Enable GPU awareness in PAMI, by adding one mca
                     option and also by remotely setting LD_PRELOAD of
                     libpami_cudahook.so.
EOF
    }

# ------------------------------------------------
# -------- -entry=
    @entry_libs = ();
    @entry_baselibs = ();
    $entry_verbose = 0;

    if ($help) { print << 'EOF'

  [Dynamic MPI Profiling interface]
    -entry <lib>,.. : list of PMPI wrapping libraries.  Each <lib> can be
                      of the form libfoo.so, /path/to/libfoo.so, or just
                      foo, which will be automatically expanded into
                      libfoo.so for simple strings consisting only of the
                      characters [a-zA-Z0-9_-]  (expansion not applicable
                      for "fort", "fortran", "v", and "vv")
    -entry fort     : included in a list of <lib> above, this indicates what
                      layer to install the base MPI product's fortran
                      calls which minimally wrap the C calls (by default
                      this is put at the top)
    -entry fortran  : same as fort
    -entrybase <lib>,.. : optionally specify what library(s) to get the
                      bottom level MPI calls from, by default RTLD_NEXT
                      which would be the libmpi the executable is linked
                      against.
    -baseentry      : same as -entrybase
    -entry v        : verbose (show the layering of the MPI entrypoints)
    -entry vv       : more verbose - the difference is 'v' shows what levels
                      are intended to be used, 'vv' happens further inside
                      the library and confirms what libraries are being opened.
                      'vv' output is less readable, but more visibly confirms
                      interception is taking place.
    -entry mpe      : turn on the pre-built MPE logging library (version
                      mpe2-2.4.9b) from Argonne National Laboratory. The
                      output .clog2 file is viewable with jumpshot.
EOF
    }

# ------------------------------------------------
# -------- -show -onlyshow -showonly (to see the modified mpirun cmdline)
    $show = 0;
    $run = 1;
    $show_syntax = "sh"; # can be sh, csh, or keyval
    $env_output_file = "";
    $env_output_file_fh_isopen = 0;

    if ($help) { print << 'EOF'

  [Help options]
    -show          : display the resulting modified mpirun command line
                     as well as run the resulting mpirun command
    -showonly      : like -show but only displays the result,
                     doesn't run the resulting mpirun command
    -onlyshow      : same as -showonly
    -show_as <syntax>    : specifies what syntax to output environment settings
                           where <syntax> can be
        sh               : write env settings for the sh shell (the default)
        csh              : write env settings for the csh shell
        keyval           : write env settings in a simple VAR VALUE syntax
    -write_env <file>        : The -write_env* options are similar to -showonly
    -write_env_sh <file>     : with a corresponding -show_as <syntax> option,
    -write_env_csh <file>    : but the generated environment is written to
    -write_env_keyval <file> : <file> instead of to stdout.
    -generate_env <file>         : The -generate_env* options are all
    -generate_env_sh <file>      : equivalent to the corresponding
    -generate_env_csh <file>     : -write_env* options. Note that the base
    -generate_env_keyval <file>  : -write_env and base -generate_env use
                                   the default <syntax> of sh
    -help          : display this message
EOF
    }
}

# helper utilities
sub isarg {
    my $string = $_[0];
    if ($arg eq "-$string" || $arg eq "--$string") { return 1; }
    return 0;
}
sub isarg_plus_option {
    my $string = $_[0];
    my $optionref = $_[1];
    if ($arg =~ /^-+$string$/) {
        ${$optionref} = shift(@ARGV);
        return 1;
    }
    if ($arg =~ /^-+$string=(.*)$/) { # should deprecate this syntax eventually
        ${$optionref} = $1;
        return 1;
    }
    return 0;
}
sub isarg_plus_list {
    my $string = $_[0];
    my $listref = $_[1];
    if ($arg =~ /^-+$string$/) {
        @{$listref} = split(/,+/, shift(@ARGV));
        return 1;
    }
    if ($arg =~ /^-+$string=(.*)$/) {
        @{$listref} = split(/,+/, $1);
        return 1;
    }
    return 0;
}

# Extra entries can be added to @ARGV from env vars.
# For parsing, args are separated on spaces.
# If a user really wants a space in an arg,
# :: becomes space, and \:: becomes ::
# All other characters aren't special, so if the env var's value contains
# quotes etc those remain as quotes in the args that go to mpirun.
sub prepend_on_argv {
    my $env = $_[0];
    my $string = $ENV{"$env"};
    my @list = ();
    my @data;
    my $arg;
    if ($string ne '') {
        @data = split(/ +/, $string);
        for $arg (@data) {
            $arg =~ s/::/ /g;
            $arg =~ s/\\ /::/g;
            push(@list, $arg);
        }
    }
    unshift(@ARGV, @list);
}

# Walk the list of arguments doing
# - for recognized multi-part Open MPI args, move all the pieces into
#   @argv_passthrough
# - for Platform added options, record state variables
# - anything else that starts with a dash must be a single-part Open MPI
#   arg and goes into @argv_passthrough
# - the first thing we see without a dash (that's also not part of a
#   multi-part arg) must be the executable so we stop parsing args there

sub process_argv {
    # allow MPIRUN_OPTIONS, MPIRUN_SYSTEM_OPTIONS
    #       MPIRUN_USER_OPTIONS, MPIRUN_ISV_OPTIONS
    # to add to ARGV.
    prepend_on_argv('MPIRUN_OPTIONS');
    prepend_on_argv('MPIRUN_SYSTEM_OPTIONS');
    prepend_on_argv('MPIRUN_USER_OPTIONS');
    prepend_on_argv('MPIRUN_ISV_OPTIONS');

    @argv_passthrough = ();
    while (@ARGV) {
        $arg = shift(@ARGV);
        if (grep { $arg eq $_ } @skip2) {
            $arg2 = shift(@ARGV);
            $arg3 = shift(@ARGV);
            push(@argv_passthrough, $arg, $arg2, $arg3);
            # keep track of any -mca we encounter
            if ($arg =~ /^-{1,2}g?mca$/ ) {
                $extra_env_from_cmdline{"OMPI_MCA_" . $arg2} = $arg3;
            }
        }
        elsif (grep { $arg eq $_ } @skip1) {
# There is one exception here, OMPI -help is documented as having 1 option
# after it and is thus in the @skip1 list but -help can actually have 1 or 0
# options (the only OMPI arg with a variable number of options after it)
            if (@ARGV) {
            $arg2 = shift(@ARGV);
                push(@argv_passthrough, $arg, $arg2);
                # keep track of any -x we encounter
                if ($arg =~ /^-{1,2}x$/ ) {
                    if ($arg2 =~ /^([^=]*)=(.*)/) {
                        $extra_env_from_cmdline{$1} = $2;
                    } else {
                        $extra_env_from_cmdline{$arg2} = $ENV{$arg2};
                    }
                }
            } else {
                push(@argv_passthrough, $arg, "full");
            }
        }

        # Detect Platform options:
        elsif (isarg('MXM'))    { force_pml_assign('yalla'); }
        elsif (isarg('mxm'))    { force_pml_assign('yalla'); }
        #elsif (isarg('mxm'))    { $prefer_pml = 'yalla'; }
        elsif (isarg('PAMI'))   { force_pml_assign('pami'); $use_pami=1; }
        elsif (isarg('pami'))   { force_pml_assign('pami'); $use_pami=1; }
        #elsif (isarg('pami'))   { $prefer_pml = 'pami'; }
# We've removed (D136551): -MXMC / -mxmc  : use Mellanox MXM via the PML cm and the MTL mxm
        elsif (isarg('MXMC-UNSUPPORTED'))   { force_pml_assign('cm'); $allow_cm_mtl = 'mxm'; }
        elsif (isarg('mxmc-UNSUPPORTED'))   { force_pml_assign('cm'); $allow_cm_mtl = 'mxm'; }
        #elsif (isarg('mxmc'))   { $prefer_pml = 'cm'; $prefer_cm_mtl = 'mxm'; }
        elsif (isarg('OFI'))    { force_pml_assign('cm'); $allow_cm_mtl = 'ofi'; }
        elsif (isarg('ofi'))    { force_pml_assign('cm'); $allow_cm_mtl = 'ofi'; }
        #elsif (isarg('ofi'))    { $prefer_pml = 'cm'; $prefer_cm_mtl = 'ofi'; }
        elsif (isarg('PSM'))    { force_pml_assign('cm'); $allow_cm_mtl = 'psm'; }
        elsif (isarg('psm'))    { force_pml_assign('cm'); $allow_cm_mtl = 'psm'; }
        #elsif (isarg('psm'))    { $prefer_pml = 'cm'; $prefer_cm_mtl = 'psm'; }
        elsif (isarg('PSM2'))    { force_pml_assign('cm'); $allow_cm_mtl = 'psm2'; }
        elsif (isarg('psm2'))    { force_pml_assign('cm'); $allow_cm_mtl = 'psm2'; }
        #elsif (isarg('psm2'))    { $prefer_pml = 'cm'; $prefer_cm_mtl = 'psm2'; }
        elsif (isarg('TCP'))    { force_pml_assign('ob1'); $allow_ob1_btl{'tcp'}=1; }
        elsif (isarg('tcp'))    { force_pml_assign('ob1'); $allow_ob1_btl{'tcp'}=1; }
        #elsif (isarg('tcp'))    {$prefer_pml='ob1'; $allow_ob1_btl{'tcp'}=1; }
        elsif (isarg('USNIC'))  { force_pml_assign('ob1'); $allow_ob1_btl{'usnic'}=1; }
        elsif (isarg('usnic'))  { force_pml_assign('ob1'); $allow_ob1_btl{'usnic'}=1; }
        #elsif (isarg('usnic'))  {$prefer_pml='ob1'; $allow_ob1_btl{'usnic'}=1; }
        elsif (isarg('IB'))     { force_pml_assign('ob1'); $allow_ob1_btl{'openib'}=1; }
        elsif (isarg('ib'))     { force_pml_assign('ob1'); $allow_ob1_btl{'openib'}=1; }
        #elsif (isarg('ib'))     {$prefer_pml='ob1'; $allow_ob1_btl{'openib'}=1; }
        elsif (isarg('IBV'))    { force_pml_assign('ob1'); $allow_ob1_btl{'openib'}=1; }
        elsif (isarg('ibv'))    { force_pml_assign('ob1'); $allow_ob1_btl{'openib'}=1; }
        #elsif (isarg('ibv'))    {$prefer_pml='ob1'; $allow_ob1_btl{'openib'}=1; }
        elsif (isarg('OPENIB')) { force_pml_assign('ob1'); $allow_ob1_btl{'openib'}=1; }
        elsif (isarg('openib')) { force_pml_assign('ob1'); $allow_ob1_btl{'openib'}=1; }
        #elsif (isarg('openib')) {$prefer_pml='ob1'; $allow_ob1_btl{'openib'}=1; }

        elsif (isarg('HCOLL'))  {$enable_hcoll=1; $hcoll_priority = 98; }
        elsif (isarg('hcoll'))  {$enable_hcoll=1; $hcoll_priority = 90; }
        elsif (isarg('FCA'))    {$enable_hcoll=1; $hcoll_priority = 98; }
        elsif (isarg('fca'))    {$enable_hcoll=1; $hcoll_priority = 90; }

        elsif (isarg_plus_option('show_as', \$tmp)) {
            if ($tmp =~ /^s/) { $show_syntax = "sh"; }
            if ($tmp =~ /^c/) { $show_syntax = "csh"; }
            if ($tmp =~ /^k/) { $show_syntax = "keyval"; }
        }
        elsif (isarg_plus_option('write_env', \$env_output_file)) {
            # use $show_syntax as whatever it's current value is
            $run = 0;
        }
        elsif (isarg_plus_option('write_env_sh', \$env_output_file)) {
            $show_syntax = "sh";
            $run = 0;
        }
        elsif (isarg_plus_option('write_env_csh', \$env_output_file)) {
            $show_syntax = "csh";
            $run = 0;
        }
        elsif (isarg_plus_option('write_env_keyval', \$env_output_file)) {
            $show_syntax = "keyval";
            $run = 0;
        }
        elsif (isarg_plus_option('generate_env', \$env_output_file)) {
            # use $show_syntax as whatever it's current value is
        }
        elsif (isarg_plus_option('generate_env_sh', \$env_output_file)) {
            $show_syntax = "sh";
            $run = 0;
        }
        elsif (isarg_plus_option('generate_env_csh', \$env_output_file)) {
            $show_syntax = "csh";
            $run = 0;
        }
        elsif (isarg_plus_option('generate_env_keyval', \$env_output_file)) {
            $show_syntax = "keyval";
            $run = 0;
        }
        # to specify on-host communication method,
        # use -intra=<val> or -intra <val>
        elsif (isarg_plus_list('hostlist', \@list)) {
            for $tmp (@list) {
                if ($tmp =~ /^(.+):([0-9]+)$/) {
                    for ($i=0; $i<$2; ++$i) {
                        push @hostlist, $1;
                    }
                } else {
                    push @hostlist, $tmp;
                }
            }
        }
        elsif (isarg_plus_list('intra', \@list)) {
            # allow -intra=nic,vader,shm (where shm==vader)
            # none of these force the pml to be ob1, but they are only
            # active if the pml is ob1.
            for $arg2 (@list) {
                if ($arg2 eq 'nic')   { $allow_ob1_btl_forshmem{'nic'}=1; }
                if ($arg2 eq 'vader') { $allow_ob1_btl_forshmem{'vader'}=1; }
                if ($arg2 eq 'shm')   { $allow_ob1_btl_forshmem{'vader'}=1; }
            }
        }
        elsif (isarg('prot')) { $use_prot = 1; }
        elsif (isarg('protlazy')) { $use_prot = 2; }
        elsif (isarg('lazyprot')) { $use_prot = 2; }
        elsif (isarg('pami_noib')) { $use_pami_noib = 1; $use_pami=1; }
        elsif (isarg('pami_ft')) { $use_pami_ft = 1; $use_pami=1;}
        elsif (isarg_plus_list('verbsbypass', \@list)) {
            # allow 3.3.1, 3.4.0, 4.0.0, 4.1.4, auto or off
          if (0 != $use_pami_verbsbypass) {
            # If already set to off, don't parse again.
            $use_pami=1;
            for $arg2 (@list) {
                if ($arg2 =~ /^off/i)   { $use_pami_verbsbypass = 0; }
                elsif ($arg2 =~ /^3\.3/)   { $use_pami_verbsbypass = 331; }
                elsif ($arg2 =~ /^3\.4/)   { $use_pami_verbsbypass = 341; }
                elsif ($arg2 =~ /^4\.0/)   { $use_pami_verbsbypass = 404; }
                elsif ($arg2 =~ /^4\.1/)   { $use_pami_verbsbypass = 414; }
	        elsif ($arg2 =~ /^auto/i)  {
		     # mlx5_core for connect IB adapter cards
		     # mlx4_core for connectX-3 adapter cards
		     my $mof_version='/sys/module/mlx5_core/version';
		     unless(-e $mof_version){
		          $mof_version='/sys/module/mlx4_core/version';
	             }
		     #for future releases we can add up to the unless/if ladder of $mof_version
		     if(open(my $fh, '<:encoding(UTF-8)', $mof_version)) {
		          my $row=<$fh>;
		          my $exclude=0;
		          my $version_number= (split /-/,$row)[0];
		          %inclusion_map=('3.3',331,'3.4',341, '4.0', 404, '4.1', 414);
		          @exclusion_list=('3.2-*.*.*','3.1-3.*.*');
		          while(@exclusion_list){
		               $e_element=shift(@exclusion_list);
			       if($row=~ $e_element){
			            $exclude=1;
			       }
		          }
		          if(exists($inclusion_map{"$version_number"}) && !($exclude))   {
			       $use_pami_verbsbypass = $inclusion_map{"$version_number"};
		          }
		          else{
			       $use_pami_verbsbypass=0;
		          }
		          if($use_pami_verbsbypass == 0) {
			       print "verbsbypass was requested but cannot be enabled due to unsupported Mellanox OFED level \n";
		          }
		     }
		     else{
		          $use_pami_verbsbypass = 0;
		          print "verbsbypass was requested but cannot be enabled due to unsupported Mellanox OFED level \n";
		     }
	        }
                else {
                     print "Error: bad argument passed to -verbsbypass $arg2.\n";
                     print "       Valid arguments are 4.1, auto or off\n";
                     exit 1;
                }
            }
          }
        }
        elsif (isarg_plus_list('stdio', \@list)) {
            # allow -stdio=p,i[+|-|all|none|rank#],file:prefix
            for $arg2 (@list) {
                if ($arg2 eq 'p')   { $stdio_tag_output = 1; }
                if ($arg2 eq 't')   { $stdio_timestamp_output = 1; }
                if ($arg2 =~ /i(.*)/) {
                    # default to all if 'i' occurs,
                    # detect - / none / <rank-number>
                    $stdin_goes_to = 'all';
                    if ($1 eq 'none' || $1 eq '-') {
                        $stdin_goes_to = 'none';
                    }
                    elsif ($1 =~ /^([0-9]+)$/) {
                        $stdin_goes_to = $1;
                    }
                }
                if ($arg2 =~ /file[s]?:(.+)/) {
                    $output_filename = $1;
                }
                if ($arg2 =~ /file[s]?$/) {
                    $output_filename = 'mpi_output';
                }
            }
        }
        elsif (isarg_plus_list('restrict_libs', \@list)) {
            $arg2 = join(",", @list);
            $restrict_libs = $arg2;
        }
        elsif (isarg_plus_list('netaddr', \@list)) {
            # this option is actually -netaddr rank:str,str now
            # so the list is wrongly separated above
            $arg2 = join(",", @list);

            # <specs> or <type>:<specs>
            if ($arg2 =~ /([a-z]+):(.*)/) {
                if ($1 eq 'control') { $netaddr_mpirun = $2; }
                if ($1 eq 'mpirun')  { $netaddr_mpirun = $2; }
                if ($1 eq 'rank')    { $netaddr_rank = $2; }
            } else {
                $netaddr = $arg2;
            }
        }
        elsif (isarg_plus_list('aff', \@list)) {
            $use_aff = 1;
            for $arg2 (@list) {
                $arg2m = $arg2;

                # translations:
                if ($arg2 eq 'off')       { $arg2m = 'unbind'; }
                if ($arg2 eq 'none')      { $arg2m = 'unbind'; }
                if ($arg2 eq 'on')        { $arg2m = 'bandwidth'; }
                if ($arg2 eq 'auto')      { $arg2m = 'bandwidth'; }
                if ($arg2 eq 'automatic') { $arg2m = 'bandwidth'; }
                if ($arg2 eq 'default')   { $arg2m = 'bandwidth'; }
                if ($arg2 eq 'custom')    { $arg2m = 'custom'; }

                if ($arg2m eq 'v') { $aff_v = 1; }
                elsif ($arg2m eq 'vv') { $aff_v = 1; }
                elsif ($arg2m eq 'custom') { $use_aff = 0; }
                elsif ($arg2m eq 'unbind') { $bind_to = 'none'; }
                elsif ($arg2m =~ /width:(.*)/) { $bind_to = $1; }
                elsif ($arg2m =~ /cycle:(.*)/) {
                    $map_by = $1;
                    $rank_by = $1;
                }
                elsif ($arg2m eq 'bandwidth') {
                    $map_by = 'socket';
                    $rank_by = $bind_to;
                }
                elsif ($arg2m eq 'latency') {
                    $map_by = $bind_to;
                    $rank_by = $bind_to;
                }
                else {
                    print "WARNING: -aff <option> \"$arg2m\" unrecognized\n";
                }
            }
        }
        elsif (isarg('gpu')) { $use_gpu = 1; }
        elsif (isarg('disable_gpu_hooks')) { $disable_gpu_hooks = 1; }
        elsif (isarg_plus_list('entry', \@list)) {
# Accept special options v, and vv
# 'fort' is special too, but it is treated normally here and handled
# as special inside the library.
            for $tmp (@list) {
                if ($tmp eq 'v') { $entry_verbose = 1; }
                elsif ($tmp eq 'vv') { $entry_verbose = 2; }
                else {
                    push(@entry_libs, $tmp);
                }
            }
        }
        elsif (isarg_plus_list('entrybase', \@list)) {
            @entry_baselibs = @list;
        }
        elsif (isarg_plus_list('baseentry', \@list)) {
            @entry_baselibs = @list;
        }
        elsif (isarg('show')) { $show = 1; }
        elsif (isarg('onlyshow')) { $show = 1; $run = 0; }
        elsif (isarg('showonly')) { $show = 1; $run = 0; }
        elsif (isarg('jsrun')) {
# For this arg we want to consume the entire remaining @ARGV for jsrun
            $use_jsrun = 1;
            while (@ARGV) {
                push @jsrun_args, shift(@ARGV);
            }
        }
        elsif ($arg =~ /^-/) {
            push(@argv_passthrough, $arg);
        }
        else {
            # We must have reached the executable, everything beyond this
            # point is passthrough
            push(@argv_passthrough, $arg);
            while (@ARGV) {
                $arg = shift(@ARGV);
                push(@argv_passthrough, $arg);
            }
        }
    }

    # If no "Interconnect selection" passed, then default to -PAMI
    # Note: don't set $use_pami here because user didn't specify.
    if ($force_pml eq '') {
        force_pml_assign('pami');
    }
}

sub force_pml_assign {
    if ($force_pml eq '') {
        $force_pml = $_[0];
    } else {
        # Print List is from [Interconnect selection]
        if ($uname_m eq 'x86_64') {
            print "WARNING: Interconnect selection options -MXM, -PAMI, -PSM, -TCP, -IBV, -USNIC are mutually exclusive.\n";
        } else {
            print "WARNING: Interconnect selection options -MXM, -PAMI, -TCP, -IBV are mutually exclusive.\n";
        }
        print "         Only the first option is in effect.\n";
    }
}

sub display_gpu_and_not_pami_error_msg() {
    print "-"x70 . "\n";
    print "Error: Incompatible arguments supplied.\n";
    print "       The -gpu option must be supplied with either the -PAMI / -pami\n";
    print "       option or with none of the options from the section on\n";
    print "       \"Interconnect selection\" specified on the same command line.\n";
    print "-"x70 . "\n";
    exit -1;
}

# display_incompatible_args_error_msg(ARG1, ARG2)
# - A generic Error: "X" is incompat with "Y"
#
sub display_incompatible_args_error_msg {
    die "display_incompatible_args_error_msg() requires exactly two args." if (@_ != 2);

    print "-"x70 . "\n";
    print "Error: Incompatible arguments supplied.\n";
    print "       The $_[0] option CANNOT be used with $_[1] option.\n";
    print "-"x70 . "\n";
    exit -1;
}

# Use state vars to add more Open MPI env var settings via record_env() and
# more command line args in @argv_additions. The preference is toward
# record_env() when possible.
#
# NOTE: Probably not safe to call multiple times, but we're not.
sub produce_new_arguments {
    @argv_additions = ();

# @hostlist
    $using_hostlist = 0;
    if (scalar(@hostlist) >= 2) {
        $using_hostlist = 1;

        # Modern system limits are pretty generous, but I'm still switching to a
        # hostfile if the hostlist reaches 1k.

        $str = join(",", @hostlist);
        if (length($str) < 1024) {
            push(@argv_additions, "-host", $str,
                "--gmca", "rmaps", "seq");
        } else {
            ($fh_hostlist, $filename_hostlist) = mk_tmp_file();
            push(@argv_additions, "-hostfile", $filename_hostlist,
                "--gmca", "rmaps", "seq");
            for $host (@hostlist) {
                print $fh_hostlist "$host\n";
            }
            close $fh_hostlist;
        }
    }

# $force_pml
# can specify which pml (yalla / pami / cm / ob1) to use
    if ($force_pml ne '') {
        if( $force_pml ne 'pami' && 1 == $use_gpu ) {
            display_gpu_and_not_pami_error_msg();
        }
        record_env('OMPI_MCA_pml', $force_pml);
        if ($force_pml ne 'pami') {
            record_env('OMPI_MCA_osc', '^pami');
        } else {
            record_env("OMPI_MCA_osc", 'pami');
            # when using PAMI PML, only init self btl
            record_env("OMPI_MCA_btl", 'self');
        }
    }
# $prefer_pml
# can spcify which pml should be at the top of the stack
# Note: $force_pml overrides this option
# Default priorities of PML and MTL components:
#  60 : pami
#  50 : yalla (also known as MXM)
#  20 : ob1
#   5 : ucx
#  -1 : cm (set by MTL) - does not have a pml_cm_priority MCA parameter
#                       - priority taken from the MTL component selected
#       40 : psm2
#       30 : psm
#       30 : mxm (also known as MXMC)
#       25 : ofi
#       10 : portals4
#  -1 : v
    elsif ($prefer_pml ne '' && $prefer_pml ne 'cm' ) {
        if( $prefer_pml ne 'pami' && 1 == $use_gpu ) {
            display_gpu_and_not_pami_error_msg();
        }
        record_env("OMPI_MCA_pml_".$prefer_pml."_priority", 100);
    }
# $allow_cm_mtl
# IF pml==cm, this can specify which mtl for cm to use.
# (mxm / ofi / psm).  It doesn't force pml=cm though.
    if ($allow_cm_mtl ne '') {
        if( 1 == $use_gpu ) {
            display_gpu_and_not_pami_error_msg();
        }
        # If we are preferring 'cm' and the 'mtl' then 'cm' adpots the priority
        # of the MTL selected, so make sure that MTL has a high priority
        if( $prefer_pml eq 'cm' ) {
            record_env("OMPI_MCA_mtl_".$allow_cm_mtl."_priority", 100);
        }
        record_env("OMPI_MCA_mtl", $allow_cm_mtl);
    }
# $prefer_cm_mtl
# can specify which mtl should be at the top of the stack
# Note: $allow_cm_mtl overrides this option
# See above for default priorities
    elsif ($prefer_cm_mtl ne '') {
        if( 1 == $use_gpu ) {
            display_gpu_and_not_pami_error_msg();
        }
        # Put this MTL at the top of the MTL stack
        record_env("OMPI_MCA_mtl_".$prefer_cm_mtl."_priority", 100);
    }
# enable_hcoll
# hcoll_priority
    if ($enable_hcoll) {
        if( 1 == $use_gpu ) {
            display_gpu_and_not_pami_error_msg();
        }
        if ( 1 == $use_prot) {
            print "-HCOLL not compatible with -prot, changing -prot to -protlazy.  Continuing...\n";
            $use_prot = 2;
        }
        record_env("OMPI_MCA_coll_hcoll_enable", $enable_hcoll);
        record_env("OMPI_MCA_coll_hcoll_priority", $hcoll_priority);
        record_env("OMPI_MCA_pml_pami_context_advance_niter", "1");

        if ($force_pml eq 'pami') {
            # Some bad interaction between PAMI and HCOLL with HCOLL's reduce function
            # Work around for now is to disable HCOLL's collective
            # https://jazz07.rchland.ibm.com:21443/jazz/resource/itemName/com.ibm.team.workitem.WorkItem/157704
            my $skip_env_set = 0;
            # Make sure they have not overridden the value in the environment
            if( $ENV{'HCOLL_ML_DISABLE_REDUCE'} ) {
                $skip_env_set = 1;
            }
            else {
                # Make sure they have not overridden the value on the command line (-x option)
                foreach $pt (@argv_passthrough) {
                    if( $pt =~ /^HCOLL_ML_DISABLE_REDUCE/ ) {
                        $skip_env_set = 1;
                        last;
                    }
                }
            }
            # Ok, they have not overridden the environment variable, so set it
            if( 0 == $skip_env_set ) {
                record_env("HCOLL_ML_DISABLE_REDUCE", "1");
            }
        }
    }

# $allow_ob1_btl
# $allow_ob1_btl_forshmem
#
# These contain the user-supplied BTLs, where an empty list of keys means
# unspecified. The on/off host lists are separate and need to be
# combined in a logical way here.  The on-host "BTL" of "nic" means
# don't put vader in the list.
    if (%allow_ob1_btl || %allow_ob1_btl_forshmem) {
        # something is being specified for BTLs, so all must be specified
        @list = ('self');

        # if 'nic' is specified, put nothing in the list for on-host
        # else if anything is specified, put that
        # else put all the on-host BTLs, eg the default if unspecified
        if ($allow_ob1_btl_forshmem{'nic'}) {
        }
        elsif (%allow_ob1_btl_forshmem) {
            push(@list, keys(%allow_ob1_btl_forshmem));
        }
        else {
            push(@list, @all_btls_onhost);
        }

        # for off-host, either use the specified list or all-off-host
        # if unspecified
        if (%allow_ob1_btl) {
            push(@list, keys(%allow_ob1_btl));
        }
        elsif ((keys %allow_ob1_btl_forshmem) eq 0) {
            push(@list, @all_btls_offhost);
        }

        record_env("OMPI_MCA_btl", join(",", @list));
    }

# $use_prot
# 1 = regular prot in init, 2 = lazy prot in finalize
    if ($use_prot) {
        record_env("OMPI_MCA_platform_prot", $use_prot);
    }

# $stdio_tag_output
# $stdio_timestamp_output
# $stdin_goes_to
# $output_filename
    if ($stdio_tag_output) {
        record_env("OMPI_MCA_orte_tag_output", "1");
    }
    if ($stdio_timestamp_output) {
        record_env("OMPI_MCA_orte_timestamp_output", "1");
    }
    if ($stdin_goes_to ne '') {
        push(@argv_additions, "-stdin", $stdin_goes_to);
    }
    if ($output_filename ne '') {
        record_env("OMPI_MCA_orte_output_filename", $output_filename);
    }

# $restrict_libs
    if ($restrict_libs ne '') {
        record_env("OMPI_MCA_mpi_restrict_libs", $restrict_libs);
    }

# $netaddr
# $netaddr_mpirun
# $netaddr_rank
    if ($netaddr ne '') {
        record_env("OMPI_MCA_btl_tcp_if_include", $netaddr);
        record_env("OMPI_MCA_btl_usnic_if_include", $netaddr);
        record_env("OMPI_MCA_oob_tcp_if_include", $netaddr);
    }
    if ($netaddr_mpirun ne '') {
        record_env("OMPI_MCA_oob_tcp_if_include", $netaddr_mpirun);
    }
    if ($netaddr_rank ne '') {
        record_env("OMPI_MCA_btl_tcp_if_include", $netaddr_rank);
        record_env("OMPI_MCA_btl_usnic_if_include", $netaddr_rank);
    }

# $use_aff
# $aff_v
# $bind_to
# $map_by
# $rank_by
    if ($aff_v) {
        # --report-bindings
        record_env("OMPI_MCA_hwloc_base_report_bindings", "1");
    }
    if ($use_aff) {
        if ($bind_to ne 'none') {
            # --bind-to $bind_to --map-by $map_by --rank-by $rank_by
            record_env("OMPI_MCA_hwloc_base_binding_policy", $bind_to);
            record_env("OMPI_MCA_rmaps_base_mapping_policy", $map_by);
            record_env("OMPI_MCA_rmaps_base_ranking_policy", $rank_by);
        } else {
            # --bind-to none
            record_env("OMPI_MCA_hwloc_base_binding_policy", "none");
        }
    }

# $use_gpu
# Only do MPI gpu buffer checks if they pass -gpu
    if ($use_gpu) {
        record_env("OMPI_MCA_pml_pami_enable_cuda", "1");
    }

# Always LD_PRELOAD libpami_cudahook even without -gpu
    if (! $disable_gpu_hooks) {
        record_env('OMPI_LD_PRELOAD_POSTPEND_DISTRO',
            "$ENV{OPAL_LIBDIR}/libpami_cudahook.so");
    } else {
        if(1 == $use_gpu) {
            display_incompatible_args_error_msg("-gpu", "-disable_gpu_hooks");
        }
        record_env("PAMI_DISABLE_CUDA_HOOK", "1");
    }

# pami_dir - what pami would we use if we're using pami?
    # First read default pami from smpi.conf: OMPI_MCA_schizo_ompi_prepend_ld_library_path
    $pami_dir = $ENV{'OMPI_MCA_schizo_ompi_prepend_ld_library_path'};
    chomp $pami_dir;
    if (0 < $use_pami) {
        if (0 < $use_pami_verbsbypass) {
            # use_pami_verbsbypass_ver is valid 3digit: 414
            $pami_dir = "pami_$use_pami_verbsbypass";
            if( 0 < $use_pami_ft ) { $pami_dir .= "_ft"; }
        } elsif (0 < $use_pami_noib) {
            $pami_dir = "pami_noib";
            if( 0 < $use_pami_ft ) { 
                display_incompatible_args_error_msg("-pami_noib","-pami_ft"); 
            }
        } elsif (0 < $use_pami) {
            $pami_dir = "pami_port";
            if( 0 < $use_pami_ft ) { $pami_dir .= "_ft"; }
        }
    }

# Add a dir for finding PAMI and LIBCOLLECTIVES library.
    if (0 < $use_pami || $force_pml eq "pami" ) {
       $tmp_dir = "$pami_dir";
    } else {
       $tmp_dir = "mpicoll";
    }
    # postpend HCOLL dirs here if requested.
    if ($enable_hcoll) {
        # postpend the HCOLL lib dirs:
        $tmp_dir = "$tmp_dir:/opt/mellanox/hcoll/lib:/opt/mellanox/sharp/lib";
    }

    # Now always use OMPI_MCA_schizo_ompi_prepend_ld_library_path, to find libcollectives.so
    record_env('OMPI_MCA_schizo_ompi_prepend_ld_library_path', "$tmp_dir");
    # pami_dir is a relative path, and occurs at the start of $tmpdir
    # so prefix with the absolute location of that library.
    push @jsrun_ld_library_path_additions, $ENV{'OPAL_PREFIX'}."/lib/".$tmp_dir;
    if (0 < $use_pami_verbsbypass) {
# This OMPI_ var is just for informational purposes.
# Not really used, but might be wanted downstream.
        record_env('OMPI_IBM_SPECTRUM_MPI_USE_PAMI_VERBSBYPASS',
            "$use_pami_verbsbypass");
    }

    record_env('LD_LIBRARY_PATH');

# @entry_libs
# @entry_baselibs
# $entry_verbose
    @entry_libs = syntax_convenience_entry(@entry_libs);
    @entry_baselibs = syntax_convenience_entry(@entry_baselibs);
    $entry_is_used = 0;
    if (@entry_libs) {
        record_env("OMPI_MCA_entry_libs", join(",", @entry_libs));
        $entry_is_used = 1;
    }
    if (@entry_baselibs) {
        record_env("OMPI_MCA_entry_baselibs", join(",", @entry_baselibs));
        $entry_is_used = 1;
    }
    if ($entry_is_used) {
# It would be nice to just prepend LD_LIBRARY_PATH here, but Open MPI
# has too many separate paths where it prepends OPAL_PREFIX/lib into
# LD_LIBRARY_PATH which would negate this. An easier but uglier approach
# is a complete reset of OPAL_PREFIX.

        $tmp = $ENV{'OPAL_PREFIX'};
        if ( -d "$tmp/profilesupport" &&
             -x "$tmp/profilesupport/bin/orterun")
        {

            $ENV{'OPAL_PREFIX'} = "$tmp/profilesupport";
            $ENV{'MPI_ROOT'} = "$tmp/profilesupport";
            record_env('OPAL_PREFIX');
            record_env('MPI_ROOT');
        }

# but don't rely completely on another code path to prepend the above
# OPAL_PREFIX/lib, ensure the new one is there:
        $tmp = $ENV{'LD_LIBRARY_PATH'};
        if ("$tmp" eq '') {
            $ENV{'LD_LIBRARY_PATH'} = "$ENV{'OPAL_PREFIX'}/lib/profilesupport";
        } else {
            $ENV{'LD_LIBRARY_PATH'} =
                "$ENV{'OPAL_PREFIX'}/lib/profilesupport:$tmp";
        }
        push @jsrun_ld_library_path_additions, "$ENV{'OPAL_PREFIX'}/lib/profilesupport";
        record_env('LD_LIBRARY_PATH');
    }
    if ($entry_verbose) {
        record_env("OMPI_MCA_entry_verbose", $entry_verbose);

        print "Entrypoint MPI wrapper levels:\n";
        $i = 1;
        if (!grep { $_ eq 'fort' || $_ eq 'fortran' } @entry_libs) {
            if (@entry_libs) {
                print "  $i. (fortran from base product)\n";
                ++$i;
            }
        }
        while (@entry_libs) {
            $tmp = shift(@entry_libs);
            if ($tmp eq 'fort' || $tmp eq 'fortran') {
                # if 'fort' is the last item in the list, that gets ignored
                if (@entry_libs) {
                    print "  $i. (fortran from base product)\n";
                    ++$i;
                }
            } else {
                print "  $i. $tmp\n";
                ++$i;
            }
        }
        print "  $i. base product\n";
        print "Entrypoint MPI base product:\n";
        if (! @entry_baselibs) {
            print "  (default MPI: libmpi_ibm.so,libmpi_mpifh.so)\n";
        } else {
            $i = 1;
            for $tmp (@entry_baselibs) {
                print "  $i. $tmp\n";
                ++$i;
            }
        }
    }

    if(0 < $use_pami_verbsbypass) {
       if ($force_pml eq '') {
           # FIXME: I think $force_pml is already assigned in this subroutine,
           # and calling force_pml_assign is too late here!
           # Just remove this after more TESTing
           force_pml_assign('pami');
       } elsif ($force_pml ne 'pami') {
           print "Error: -verbsbypass <arg> requires PAMI PML to run.\n";
           exit 1;
       }
    }
    if(0 < $use_pami_noib) {
       if(0 < $use_pami_verbsbypass) {
           print "Error: -verbsbypass <arg> and -pami_noib are incompatible arguments.\n";
           exit 1;
       }
       if ($force_pml eq '') {
           # FIXME: I think $force_pml is already assigned in this subroutine,
           # and calling force_pml_assign is too late here!
           # Just remove this after more TESTing.
           force_pml_assign('pami');
       } elsif ($force_pml ne 'pami') {
           print "Error: -pami_noib requires PAMI PML to run.\n";
           exit 1;
       }
    }

# Runtime checing for LSF environment: if no LSF cluster is found we disable related LSF
# module initialization. This this the fix for STG Defect 102466.
    if ((! defined $ENV{'LSF_ENVDIR'}) || (! -d $ENV{'LSF_ENVDIR'})) {
        record_env("OMPI_MCA_ess", "^lsf");
        record_env("OMPI_MCA_plm", "^lsf");
        record_env("OMPI_MCA_ras", "^lsf");
    }
}

# As a convenience feature allow a shortcut of "-entry foo" to mean
# "-entry libfoo.so". This substitution is only done for simple strings
# containing only a-zA-Z0-9_- and also isn't done for any of the
# already-reserved special -entry strings (v / vv / fort / fortran) since
# those wouldn't be present in the list at this point.

sub
syntax_convenience_entry {
    my @in = @_;
    my @out = @_;
    my $item;
    @out = ();
    for $item (@in) {
        if ($item =~ /^[a-zA-Z0-9_-]*$/
            && $item ne 'fort' && $item ne 'fortran')
        {
            $item = "lib${item}.so";
        }
        push @out, $item;
    }
    return @out;
}

sub
set_extra_env_for_jsrun {
    my $key;
    my @keys;

    # Setup some environment variables we need to pass along for JSM
    # 1) Setup OMPI_LD_LIBRARY_PATH_PREPEND_DISTRO from LD_LIBRARY_PATH parts
    $ENV{OMPI_LD_LIBRARY_PATH_PREPEND_DISTRO} = join(":", @jsrun_ld_library_path_additions);
    record_env("OMPI_LD_LIBRARY_PATH_PREPEND_DISTRO");

    # 2) convert --mca / --gmca / -x to envars from what we parsed
    # earlier in the incoming "mpirun ..." cmdline. any -mca / -x we saw
    # were put into a hash %extra_env_from_cmdline. For the usual
    # run-mode where we exec "mpirun ..." this var is unused, but for jsrun
    # we'll create an env setting for every -mca / -x we saw on the cmdline.

    @keys = keys(%extra_env_from_cmdline);
    for $key (@keys) {
        $ENV{$key} = $extra_env_from_cmdline{$key};
        record_env($key);
    }

    $ENV{SMPI_INTERNAL_MPIRUN_LAUNCH_JSM} = 1;
    record_env('SMPI_INTERNAL_MPIRUN_LAUNCH_JSM');
}

# The context of this function is that we started with an SMPI mpirun command
# which we've converted into a combination of environment settings and generated
# a new OMPI mpirun command line (@cmd). But this function is called when the
# command was "mpirun .. -jsrun .." and we're about to generate a jsrun command
# instead of an OMPI mpirun command. Ideally in this case the current @cmd has
# been emptied of all its arguments other than -mca var val and -x var.
# But if it hasn't, that should be a warning or error.
sub
check_for_unrecognized_mpirun_args {
    my @args = @{$_[0]};
    my $is_fatal = $_[1];
    my $what = $_[2];
    my $is_good_to_run;
    my $arg;
    my $warn_or_error = "warning";

    if ($is_fatal) { $warn_or_error = "error"; }

    # warn or error if we ended up with mpirun args that won't make
    # sense to jsrun
    $is_good_to_run = 1;
    while (@args) {
        $arg = shift(@args);
        if ($arg =~ /^-{1,2}g?mca$/ ) {
            shift(@args);
            shift(@args);
        } elsif ($arg =~ /^-{1,2}x$/ ) {
            shift(@args);
        } else {
            print STDERR "$warn_or_error, ignoring $what: $arg\n";
            $is_good_to_run = 0;
        }
    }
# If we encountered a parameter that we don't support translating then give up.
    if ($is_fatal && !$is_good_to_run) {
        exit(2);
    }
}

sub
display_env_setting {
    my $var = $_[0];;
    my $val;
    my $line;

    $val = $ENV{$var};

    if ($show_syntax eq 'sh') {
        $line = "export $var=" . myescape($val);
    } elsif ($show_syntax eq 'csh') {
        $line = "setenv $var " . myescape($val);
    } elsif ($show_syntax eq 'keyval') {
        $line = "$var $val";
    }

    if ($show) {
        print "$line\n";
    }
    if ($env_output_file ne '') {
        if (! $env_output_file_fh_isopen) {
            open($env_output_file_fh, "> $env_output_file");
            $env_output_file_fh_isopen = 1;
        }
        print $env_output_file_fh "$line\n";
    }
}

# Create a tmp file in $MPI_TMPDIR or /tmp , also delete old files
# with the prefix we're using.
#
# return (file handle, filename)
sub
mk_tmp_file {
    my $fh;
    my $filename;
    my $whoami;
    my $hostname;
    my $dir;
    my $e;
    my $pid;

    $whoami = `whoami`;     chomp $whoami;
    $hostname = `hostname`; chomp $hostname;
    if (exists $ENV{'MPI_TMPDIR'}) {
        $dir = $ENV{'MPI_TMPDIR'};
    } else {
        $dir = "/tmp";
    }

    # We want to exec into the real mpirun later, and tempfile()'s automatic
    # deletion of the file doesn't do anything after an exec. So we need some
    # cleanup code of our own, and we use UNLINK 0 in the later tempfile()
    # so it won't even try to cleanup.
    #
    # rm files with our user/hostname if the pid doesn't exist.

    opendir $fh, "$dir";
    while ($e = readdir($fh)) {
        if ($e eq $filename) { next; }
        if ($e !~ /^\.mpirun\.hostlist\.${whoami}\.${hostname}\.([1-9][0-9]*)\./) {
            next;
        }
        $pid = $1;

        if (!kill(0, $pid)) {
            unlink "$dir/$e";
        }
    }
    closedir $fh;

    ($fh, $filename) = tempfile(
        ".mpirun.hostlist.${whoami}.${hostname}.$$.XXXXXXXX",
        DIR => $dir, UNLINK => 0);

    return ($fh, "$filename");
}


# Basic VAR = VAL syntax from a .conf file
# returns a hash
# If you want to read multiple conf files and have them building into a single
# hash, a hash reference can be passed as arg2, and it will be the starting pt
#
# A special hash entry $hash{":order:"}{VAR} gives an index number in case you
# want to know what order the entries appeared (foo=x bar=y foo=z would have
# foo showing the later index).
#
# Syntax features:
#   VAR = VAL
#   VAR += VAL                     : append to VAR
#   unset VAR                      : remove VAR from the hash
#   spaces before/after var name   : ignored
#   spaces before/after initial =  : ignored
#   $VAR and ${VAR}                : expanded from env or prev .conf entry
#   $$  or  $DOLLAR                : the single character $
#   $SPACE                         : a single space -- this could be used
#                                    if you really wanted leading spaces
#                                    in a value
#   $RETURN                        : the return character '\n'
#   blank lines                    : ignored
#   lines beginning with #         : comments, ignored
# Note: there is no syntax for line-continuation of long values, but
# multiple sets to the same variable are allowed, as is VAR += VAL.
# for example
#   FOO = 123
#   FOO = ${FOO}456
#   FOO += 789
# results in FOO having the value 123456789.

sub read_conf {
    my $file = $_[0];
    my $fh;
    my $line;
    my $var;
    my $vars;
    my $val;
    my $is_append;
    my $repl_var;
    my $repl_val;
    my $left;
    my $right;
    my $index;
    my %hash = ();
    my $href = $_[1];

    if ($href) {
        %hash = %$href;
    }
    $index = 0;
    if (defined $hash{":index:"}) {
        $index = $hash{":index:"}
    }

    open($fh, "< $file");
    while ($line = <$fh>) {
        chomp $line;
        $is_append = 0;
        if ($line =~ /^\s*([A-Za-z0-9_]+)\s*(\+?)=\s*(.*)/) {
            $var = $1;
            if ($2 eq "+") { $is_append = 1; }
            $val = $3;
            # All occurrences of $ are substitutions ($ followed by
            # something that isn't $ or VAR or {VAR} is invalid and
            # will just be removed)
            # %perc and %dollar will be used as special codes during
            # the translation to allow looping until all $ are gone
            $val =~ s/%/%perc/g;
            $val =~ s/\$\$/%dollar/g;
            while ($val =~ /^([^\$]*)\$({?)([A-Za-z0-9_]*)(}?)(.*)$/) {
                $left = $1;
                $repl_var = $3;
                $right = $5;
                # hard to decide what to do with ${var, unmatched { allowing it
                if ($2 eq "") { $right = $4 . $right; } # but $var} keep }
                # note that $repl_val needs to use %perc and %dollar codes too
                # since the whole $val will be decoded at the end
                $repl_val = "";
                if ($repl_var ne "") {
                    if ($repl_var eq 'DOLLAR') {
                        $repl_val = "\$";
                    } elsif ($repl_var eq 'SPACE') {
                        $repl_val = " ";
                    } elsif ($repl_var eq 'RETURN') {
                        $repl_val = "\n";
                    } elsif (defined $hash{$repl_var}) {
                        $repl_val = $hash{$repl_var};
                    } elsif (defined $ENV{$repl_var}) {
                        $repl_val = $ENV{$repl_var};
                    }
                }
                $repl_val =~ s/%/%perc/g;
                $repl_val =~ s/\$/%dollar/g;
                $val = $left . $repl_val . $right;
            }
            $val =~ s/%dollar/\$/g;
            $val =~ s/%perc/%/g;
            if ($is_append) {
                $val = $hash{$var} . $val;
            }
            $hash{$var} = $val;
            $hash{":order:"}{$var} = ++$index;
            $hash{":index:"} = $index;
#           print "$var = $val\n";
        }
        elsif ($line =~ /^\s*unset\s+(.*)/) {
            $vars = $1;
            for $var (split(/\s+/, $vars)) {
                delete $hash{$var};
#               print "unset $var\n";
            }
        }
    }
    close($fh);
    return %hash;
}

# A conf fill will be loaded in order at
#   1. /etc/smpi.conf
#   2. $MPI_ROOT/etc/smpi.conf
#   3. ${HOME}/.smpi.conf
# having multiple conf files exist is fine, the later settings overwrite
# the earlier.

sub read_smpi_confs {
    my %conf_hash = ();
    my @keys;
    %conf_hash = read_conf("/etc/smpi.conf");
    %conf_hash = read_conf("$ENV{MPI_ROOT}/etc/smpi.conf", \%conf_hash);
    %conf_hash = read_conf("$ENV{HOME}/.smpi.conf", \%conf_hash);

# Even though the ordering doesn't really matter, I figure it might
# aid in readability of the output, so saving the vars in order:

    @keys = sort { $conf_hash{":order:"}{$a}
                   <=> $conf_hash{":order:"}{$b} }
                   keys %conf_hash;
    for $key (@keys) {
        if ($key ne ":order:" && $key ne ":index:") {
            record_env($key, $conf_hash{$key});
        }
    }
}

# Every time we set an environment variable we need to keep track
# of it so we can print a list of environment variables we set, and
# also so we can produce an OMPI_MCA_mca_base_env_list_distro to list
# vars for OMPI to propagate.
#
#   record_env VAR
#   record_env VAR VALUE
#   unrecord_env VAR
#   record_env_nopropagate VAR
#
# maintains two lists:
# @record_env_list      : env vars that have been set or changed. These
#                         are tracked by calling record_env each time an
#                         env var is set/changed
# @record_env_propagate : env vars that are not only to be exported locally
#                         but also put in OMPI_MCA_mca_base_env_list_distro
#                         to be propagated to the ranks
#
# Normally vars are automatically put in the OMPI_MCA_mca-base_env_list_distro
# list if they don't start with OMPI_MCA_, but it's possible some env change
# is not intended for that, so the _nopropagate call is to remove a var from
# the special propagation list.

sub record_env {
    my $var = $_[0];
    my $val = $_[1];

    if (defined $val) {
        $ENV{$var} = $val;
    }

    if (! grep(/^$var$/, @record_env_list)) {
        push @record_env_list, $var;
    }
    if ($var !~ /^OMPI_MCA_/ && ! grep(/^$var$/, @record_env_propagate)) {
        push @record_env_propagate, $var;
    }
}
sub record_env_nopropagate {
    my $var = $_[0];

    @record_env_propagate = grep(!/^$var$/, @record_env_propagate);
}
sub unrecord_env {
    my $var = $_[0];

    @record_env_list = grep(!/^$var$/, @record_env_list);
    @record_env_propagate = grep(!/^$var$/, @record_env_propagate);
}

sub myescape {
    my $val = $_[0];
    $val =~ s/([^A-Za-z0-9_\/.,^:-])/\\$1/g;
    return $val;
}
