#ifndef __libsymm_h__
#define __libsymm_h__

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <sys/uio.h>

#ifdef __cplusplus
extern "C"
{
#endif



  typedef void*     libsymm_client_t;   /**< A PAMI Client to be passed in */
  typedef void*     libsymm_context_t;  /**< A LIBCOLL Context to be passed in  */
  typedef void*     libsymm_epdomain_t; /**< A LIBCOLL EPDomain to be passed in */
  typedef void*     libsymm_geometry_t; /**< A LIBCOLL Geometry to be passed in */
  typedef uint32_t  libsymm_endpoint_t; /**< Endpoint of a p2p operation    */
  typedef void*     libsymm_group_t;    /**< A group of endpoint list  */

  #define LIBSYMM_WORLD_GROUP ((void *) -1L) 

  /**
   * \brief LIBSYMM result status codes.
   */
  typedef enum
  {
    LIBSYMM_SUCCESS =  0, /**< Successful execution        */
    LIBSYMM_NERROR  = -1, /**< Generic error (-1)          */
    LIBSYMM_ERROR   =  1, /**< Generic error (+1)          */
    LIBSYMM_INVAL,        /**< Invalid argument            */
    LIBSYMM_UNIMPL,       /**< Function is not implemented */
    LIBSYMM_EAGAIN,       /**< Not currently available     */
    LIBSYMM_ENOMEM,       /**< Out of memory               */
    LIBSYMM_SHUTDOWN,     /**< Task has shutdown           */
    LIBSYMM_CHECK_ERRNO,  /**< Check the errno val         */
    LIBSYMM_OTHER,        /**< Other undefined error       */
    LIBSYMM_RESULT_EXT = 1000 /**< Begin extension-specific values */
  } libsymm_result_t;

  /**
   * \brief Initialize LIBSYMM using a libcoll_context, libcoll_epdomain and libcoll_geometry.
   *
   * LIBSYMM internally uses LIBCOLL for collective communication. The user needs to create
   * a LIBCOLL context, epdomain and world geometry using LIBCOLL_Init API and pass these parameters
   * to LIBSYMM. The pami_client and pami_context is needed to get task related info and for progress.
   * 
   * \note LIBSYMM_Init is a synchronizing collective call. All participating tasks will need to
   *       call into LIBSYMM_Init. LIBSYMM_Init is a thread-safe routine; however, only a single 
   *       thread can call LIBSYMM_Init. 
   * 
   * \param [in] pami_client       a PAMI client created by the user using PAMI_Client_create
   * \param [in] pami_context      a PAMI context created by the user using PAMI_Context_createv
   * \param [in] libcoll_context   a LIBCOLL context created by the user using LIBCOLL_Init
   * \param [in] libcoll_epdomain  a LIBCOLL epdomain created by the user using LIBCOLL_Init
   * \param [in] libcoll_geometry  a LIBCOLL wgeometry created by the user using LIBCOLL_Init
   * \param [in] use_threads       enable/disable threading 
   *
   * \retval ::LIBSYMM_SUCCESS  LIBSYMM was initialized successfullt
   * \retval ::LIBSYMM_ERROR    An error occured during Init.
   */

  libsymm_result_t LIBSYMM_Init(libsymm_client_t pami_client, 
                                libsymm_context_t pami_context,
                                libsymm_context_t libcoll_context, 
                                libsymm_epdomain_t libcoll_epdomain, 
                                libsymm_geometry_t libcoll_geometry, 
                                int use_threads);

  /**
   * \brief Creates a group of endpoints to be used when allocating symm memory.
   *
   * LIBSYMM internally uses LIBCOLL to create a geometry based on the eplist passed in.
   * The user has to use this API to create a group of endpoints. A LIBCOLL geometry created
   * outside of LIBSYMM is not compatible with the libsymm_group_t.
   * 
   * \note LIBSYMM_Create_group is a synchronizing collective call. All tasks in parent_group 
   *       will need to call into LIBSYMM_Create_group. LIBSYMM_Create_group is a thread-safe routine. 
   *       Tasks not being part of eplist (setting participant to 0), will get a null libsymm_group_t
   *       Also, the user is not responsible for providing a unique ID. LIBSYMM handles this internally.
   *
   * \note In multithreaded mode, if two threads create a group using the same parent group and requesting
   *       the same tasks, then there is no ordering guarantee on which threads will get the same group
   *       id. i.e. The application will need to determine the threads that belong to the same
   *       group.
   * 
   * \param [in]  eplist           an array of endpoints participating in the new group
   * \param [in]  num_eps          number of elements in the eplist array
   * \param [in]  parent_group     parent group of the group to be created
   * \param [in]  participant      indicate if the task is a participant in the group
   * \param [out] libsymm_group    opaque object initialized for the group
   *
   * \retval ::LIBSYMM_SUCCESS  libsymm_group was created successfully
   * \retval ::LIBSYMM_ERROR    An error occured during creation
   */

  libsymm_result_t LIBSYMM_Create_group(libsymm_endpoint_t* eplist, 
                                        size_t num_eps, 
                                        libsymm_group_t parent_group, 
                                        int participant, 
                                        libsymm_group_t* libsymm_group);

  /**
   * \brief Free any memory allocated inside of a libsymm_group 
   *
   * LIBSYMM_Destroy_group will free any internal resources allocated
   * during a groupcreate routine.  This may involve syncronization
   * with other tasks in the group to free the resource, and thus this call is synchronizing
   * 
   * \note LIBSYMM_Destroy_group is a non-synchronizing collective call.  
   * 
   * \param [in]  libsymm_group    handle to group to be freed 
   *
   * \retval ::LIBSYMM_SUCCESS  libsymm_group was created successfully
   * \retval ::LIBSYMM_ERROR    An error occured during creation
   */

  libsymm_result_t LIBSYMM_Destroy_group(libsymm_group_t libsymm_group);

  /**
   * \brief Allocates symmetric memory for all tasks in the libsymm_group 
   *
   * LIBSYMM_Symm_alloc will allocate a symmetric remotely accessible memory
   * of at least size "size"
   * 
   * \note LIBSYMM_Symm_alloc is a synchronizing collective call on all tasks
   * in libsymm_group
   *
   * \note In multithreaded mode, no two threads can allocate symmultaneously using
   *       the same group. In this case, two distinct groups will be required for
   *       each thread.
   * 
   * \param [in]  libsymm_group    handle to group of tasks participating in alloc
   * \param [in]  size             Specifies the minimum desired size of the symmetric memory
   * \param [out] ptr              Address of allocated symmetric memory
   *
   * \retval ::LIBSYMM_SUCCESS  Symmetric memory was successfully allocated
   * \retval ::LIBSYMM_ERROR    An error occured during allocation
   * \retval ::LIBSYMM_ENOMEM   No memory available for allocation
   */
  libsymm_result_t LIBSYMM_Symm_alloc(libsymm_group_t libsymm_group, 
                                      size_t size, 
                                      void ** ptr);

  /**
   * \brief Frees symmetric memory on the task calling the routine 
   *
   * LIBSYMM_Symm_free will release a symmetric remotely accessible memory
   * on the task calling the routine
   * 
   * \note LIBSYMM_Symm_free is a local call. It is erroneous to access memory that 
   * has been freed with LIBSYMM_Symm_free
   * 
   * \param [in] ptr            Address of allocated symmetric memory to be freed
   *
   * \retval ::LIBSYMM_SUCCESS  Symmetric memory was successfully freed 
   * \retval ::LIBSYMM_ERROR    An error occured during free
   */

  libsymm_result_t LIBSYMM_Symm_free(void *ptr);


#ifdef __cplusplus
};
#endif

#endif
